# Legobot
# Copyright (C) 2016 Brenton Briggs, Kevin McCabe, and Drew Bronson


class Message():
    """
    Messages in pykka must be passed as dictionaries. This class provides
    fields corresponding to the keys in that dictionary to facilitate the
    construction of such messages.
    """

    def __init__(self, text, metadata, should_log=False):
        self.text = text
        self.metadata = metadata
        self.should_log = should_log


class Metadata():
    """
    Message Metadata is passed as a dictionary. This class provides
    fields corresponding to the keys in that dictionary to facilitate
    the construction of such dictionaries.
    """

    def __init__(self, source, dest=None, opts=None):
        self.source = source
        self.dest = dest
        self.opts = opts



# Legobot
# Copyright (C) 2016 Brenton Briggs, Kevin McCabe, and Drew Bronson

import logging
import ssl
import threading
import time

import irc.bot
import irc.client
import irc.connection

from jaraco.stream import buffer

from Legobot.Lego import Lego
from Legobot.Message import Message, Metadata
from Legobot.Utilities import Utilities

logger = logging.getLogger(__name__)


class IgnoreErrorsBuffer(buffer.DecodingLineBuffer):
    """  Handle char decode errors better
    """
    def handle_exception(self):
        pass


irc.client.ServerConnection.buffer_class = IgnoreErrorsBuffer
irc.client.SimpleIRCClient.buffer_class = IgnoreErrorsBuffer


class IRCBot(threading.Thread, irc.bot.SingleServerIRCBot):
    """
    Create bot instance
    """
    def __init__(self,
                 baseplate,
                 channels,
                 nickname,
                 server,
                 actor_urn,
                 port=6667,
                 use_ssl=False,
                 password=None,
                 username=None,
                 ircname=None,
                 nickserv=False,
                 nickserv_pass=None,
                 rejoin_on_kick=True,
                 auto_reconnect=True):
        irc.bot.SingleServerIRCBot.__init__(self,
                                            [(server, port)],
                                            nickname,
                                            nickname)
        threading.Thread.__init__(self)

        # the obvious self.channels is already used by irc.bot
        self.my_channels = channels
        self.nickname = nickname
        self.server = server
        self.baseplate = baseplate
        self.port = port
        self.use_ssl = use_ssl
        self.password = password
        self.username = username
        self.ircname = ircname
        self.nickserv = nickserv
        self.nickserv_pass = nickserv_pass
        self.actor_urn = actor_urn
        self.rejoin_on_kick = rejoin_on_kick
        self.auto_reconnect = auto_reconnect

        self.backoff = 1

    def connect(self, *args, **kwargs):
        """
        Connect to a server.

        This overrides the function in SimpleIRCClient
        to provide SSL functionality.

        :param args:
        :param kwargs:
        :return:
        """
        if self.use_ssl:
            factory = irc.connection.Factory(wrapper=ssl.wrap_socket)
        else:
            factory = irc.connection.Factory()
        self.connection.connect(server=self.server,
                                port=self.port,
                                nickname=self.nickname,
                                connect_factory=factory,
                                password=self.password,
                                username=self.username,
                                ircname=self.ircname)

    def set_metadata(self, e):
        """
        This function sets the metadata that is common between pub and priv
        """
        metadata = Metadata(source=self.actor_urn).__dict__
        metadata['source_connector'] = 'irc'
        metadata['source_channel'] = e.target
        metadata['source_user'] = e.source
        metadata['source_username'] = e.source.split('!')[0]
        metadata['user_id'] = metadata['source_user']
        metadata['display_name'] = metadata['source_username']
        return metadata

    def on_pubmsg(self, c, e):
        """
        This function runs when the bot receives a public message.
        """
        text = e.arguments[0]
        metadata = self.set_metadata(e)
        metadata['is_private_message'] = False
        message = Message(text=text, metadata=metadata).__dict__
        self.baseplate.tell(message)

    def on_privmsg(self, c, e):
        """
        This function runs when the bot receives a private message (query).
        """
        text = e.arguments[0]
        logger.debug('{0!s}'.format(e.source))
        metadata = self.set_metadata(e)
        metadata['is_private_message'] = True
        message = Message(text=text, metadata=metadata).__dict__
        self.baseplate.tell(message)

    def on_welcome(self, c, e):
        """
        This function runs when the bot successfully connects to the IRC server
        """
        self.backoff = 1  # Assume we had a good connection. Reset backoff.
        if self.nickserv:
            if Utilities.isNotEmpty(self.nickserv_pass):
                self.identify(c, e, self.nickserv_pass)
                time.sleep(3)  # Make sure Nickserv really sees us
            else:
                logger.error('If nickserv is enabled, you must supply'
                             ' a password')

        if self.nickserv is False and self.nickserv_pass is not None:
            logger.warn('It appears you provided a nickserv password but '
                        'did not enable nickserv authentication')

        for channel in self.my_channels:
            logger.debug('Attempting to join {0!s}'.format(channel))
            c.join(channel)

    def on_kick(self, c, e):
        if self.rejoin_on_kick is True:
            time.sleep(2)
            c.join(e.target)
        return

    def on_disconnect(self, c, e):
        if self.auto_reconnect is True:
            time.sleep(2 ** self.backoff)
            try:
                self._connect()
            except:
                self.backoff += 1

    def identify(self, c, e, password):
        c.privmsg('NickServ',
                  'IDENTIFY {0!s} {1!s}'.format(self.nickname, password))
        return

    def run(self):
        """
        Run the bot in a thread.

        Implementing the IRC listener as a thread allows it to
        listen without blocking IRCLego's ability to listen
        as a pykka actor.

        :return: None
        """
        self._connect()
        super(irc.bot.SingleServerIRCBot, self).start()


class IRC(Lego):

    def __init__(self, baseplate, lock, *args, **kwargs):
        super().__init__(baseplate, lock)
        self.botThread = IRCBot(baseplate=baseplate, actor_urn=self.actor_urn,
                                *args, **kwargs)

    def on_start(self):
        self.botThread.start()

    def listening_for(self, message):
        return str(self.actor_urn) != str(message['metadata']['source'])

    def handle(self, message):
        '''
        Attempts to send a message to the specified destination in IRC
        Extends Legobot.Lego.handle()

        Args:
            message (Legobot.Message): message w/ metadata to send.
        '''

        logger.debug(message)
        if Utilities.isNotEmpty(message['metadata']['opts']):
            target = message['metadata']['opts']['target']

            for split_line in Utilities.tokenize(message['text']):
                for truncated_line in Utilities.truncate(split_line):
                    self.botThread.connection.privmsg(target, truncated_line)
                    # Delay to prevent floods
                    time.sleep(0.25)

    @staticmethod
    def get_name():
        return None



# Legobot
# Copyright (C) 2016 Brenton Briggs, Kevin McCabe, and Drew Bronson

from Legobot.Lego import Lego


class TestingConnector(Lego):
    def __init__(self, baseplate, lock, temp_file='testing_file.tmp'):
        super().__init__(baseplate, lock)
        self.temp_file = temp_file

    @staticmethod
    def listening_for(message):
        return True

    def handle(self, message):
        with open(self.temp_file, mode='w') as f:
            f.write(str(message))



# Legobot
# Copyright (C) 2017 Brenton Briggs, Kevin McCabe, and Drew Bronson


class Error(Exception):
    """Base class for exception in the Legobot module"""

    pass


class LegoError(Error):
    """Exceptions raised inside legos"""

    def __init__(self, message):
        self.message = message
        Exception.__init__(self, message)



# Legobot
# Copyright (C) 2016 Brenton Briggs, Kevin McCabe, and Drew Bronson


class Utilities():
    """
    Miscellaneous utilities for Legobot services
    """

    @staticmethod
    def tokenize(text):
        """
        Returns text split along newlines.

        Args:
            text (str): The text to be tokenized

        Returns:
            list: Text split along newlines
        """
        return text.split('\n')

    @staticmethod
    def truncate(text, length=255):
        """
        Splits the message into a list of strings of of length `length`

        Args:
            text (str): The text to be divided
            length (int, optional): The length of the chunks of text. \
                    Defaults to 255.

        Returns:
            list: Text divided into chunks of length `length`
        """

        lines = []
        i = 0
        while i < len(text) - 1:
            try:
                lines.append(text[i:i+length])
                i += length

            except IndexError as e:
                lines.append(text[i:])
        return lines

    @staticmethod
    def isNotEmpty(text):
        """
        Check if the given text is empty.

        Args:
            text (str): The text to assess

        Returns:
            bool: False if empty otherwise, True
        """

        return True if text else False



# Legobot
# Copyright (C) 2016 Brenton Briggs, Kevin McCabe, and Drew Bronson

import logging

from Legobot.Lego import Lego
from Legobot.Utilities import Utilities

logger = logging.getLogger(__name__)


class Help(Lego):
    @staticmethod
    def listening_for(message):
        if Utilities.isNotEmpty(message['text']):
            try:
                return message['text'].split()[0] == '!help'
            except Exception as e:
                logger.error(
                    'Help lego failed to check message text: {0!s}'.format(e))
                return False

    def handle(self, message):
        try:
            function = message['text'].split()[1]
        except IndexError:
            function = None

        baseplate_proxy = self.baseplate.proxy()
        legos = baseplate_proxy.children.get()

        help_str = 'No help is available. Sorry.'

        if not function:
            lego_names = []
            for lego in legos:
                lego_proxy = lego.proxy()
                if lego_proxy.get_name().get() is not None:
                    lego_names.append(lego_proxy.get_name().get())
            help_str = 'Available functions: ' + ', '.join(lego_names)

        if function:
            for lego in legos:
                lego_proxy = lego.proxy()
                if lego_proxy.get_name().get() == function:
                    try:
                        sub = message['text'].split()[2]
                        try:
                            help_str = lego_proxy.get_help(sub=sub).get()
                        except (TypeError, KeyError):
                            help_str = (function +
                                        ' has no information on ' +
                                        sub)
                    except IndexError:
                        help_str = lego_proxy.get_help().get()

        opts = self.build_reply_opts(message)

        self.reply(message, help_str, opts=opts)

    @staticmethod
    def get_name():
        return None



# Legobot
# Copyright (C) 2016 Brenton Briggs, Kevin McCabe, and Drew Bronson

import json
import logging
import threading

from pykka import ThreadingActor, ActorRegistry

from Legobot.LegoError import LegoError
from Legobot.Message import Message, Metadata

logger = logging.getLogger(__name__)


class Lego(ThreadingActor):
    class HandlerThread(threading.Thread):
        """
        This class provides a simple thread for running message handlers.
        It is used to ensure that message handlers do not block other Legos
        from running by simply taking too long to execute.
        """

        def __init__(self, handler, message):
            threading.Thread.__init__(self)
            self.handler = handler
            self.message = message

        def run(self):
            self.handler(self.message)

    def __init__(self, baseplate, lock: threading.Lock, log_file=None):
        """
        :param baseplate: the baseplate Lego, which should be \
                          the same instance of Lego for all Legos
        :param lock: a threading lock, which should be the same \
                     instance of threading.Lock for all Legos
        """
        super().__init__()
        if not lock:
            raise LegoError("Lock expected but not provided!")
        self.baseplate = baseplate
        self.children = []
        self.lock = lock
        self.log_file = log_file

    def on_receive(self, message):
        """
        Handle being informed of a message.

        This function is called whenever a Lego receives a message, as
        specified in the pykka documentation.

        Legos should not override this function.

        :param message:
        :return:
        """
        if self.log_file is not None and message['should_log']:
            message_copy = Message(message['text'],
                                   Metadata(None).__dict__,
                                   message['should_log']).__dict__
            with open(self.log_file, mode='w') as f:
                f.write(json.dumps(message_copy))
            logger.info(message['metadata']['source'])
        if self.listening_for(message):
            self_thread = self.HandlerThread(self.handle, message)
            self_thread.start()
        self.cleanup()
        for child in self.children:
            child.tell(message)

    def cleanup(self):
        """
        Clean up finished children.

        :return: None
        """
        self.lock.acquire()
        logger.debug('Acquired lock in cleanup for ' + str(self))
        self.children = [child for child in self.children if child.is_alive()]
        self.lock.release()

    def listening_for(self, message):
        """
        Return whether this Lego is listening for the provided Message.

        All Legos should override this function.

        :param message: a Message object
        :return: a boolean
        """
        return False

    def handle(self, message):
        """
        Handle the provided Message.

        All Legos should override this function.

        :param message: a Message object
        :return: None
        """
        return

    def add_child(self, child_type, *args, **kwargs):
        """
        Initialize and keep track of a child.

        :param child_type: a class inheriting from Lego to initialize \
                           an instance of
        :param args: arguments for initializing the child
        :param kwargs: keyword arguments for initializing the child
        :return:
        """
        try:
            baseplate = kwargs['baseplate']
        except:
            if self.baseplate is None:
                baseplate = self.actor_ref
            else:
                baseplate = self.baseplate
        try:
            lock = kwargs['lock']
        except:
            lock = self.lock
        child = child_type.start(baseplate, lock, *args, **kwargs)
        self.children.append(child)

    def reply(self, message, text, opts=None):
        """
        Reply to the sender of the provided message with a message \
        containing the provided text.

        :param message: the message to reply to
        :param text: the text to reply with
        :param opts: A dictionary of additional values to add to metadata
        :return: None
        """
        metadata = Metadata(source=self.actor_urn,
                            dest=message['metadata']['source']).__dict__
        metadata['opts'] = opts
        message = Message(text=text, metadata=metadata,
                          should_log=message['should_log']).__dict__
        dest_actor = ActorRegistry.get_by_urn(message['metadata']['dest'])
        if dest_actor is not None:
            dest_actor.tell(message)
        else:
            raise("Tried to send message to nonexistent actor")

    def reply_attachment(self, message, text, attachment, opts=None):
        """
        Convenience method for formatting reply as attachment (if available)
        and passing it on to the reply method. Individual connectors can then
        deal with the attachment or simply pass it on as a regular message

        :param message: the message to reply to
        :param text: the text to reply with
        :param attachment: the attachment link
        :param opts: A dictionary of additional values to add to metadata
        :return: None
        """
        if not opts:
            opts = {}

        opts['attachment'] = attachment
        opts['fallback'] = text
        text += '\n {}'.format(attachment)
        self.reply(message, text, opts)

    def build_reply_opts(self, message):
        """
        Convenience method for constructing default options for a
        reply message.

        :param message: the message to reply to
        :return: opts
        """
        try:
            source = message['metadata']['source_channel']
            thread = message['metadata'].get('thread_ts')
            opts = {'target': source, 'thread': thread}
        except LookupError:
            source = None
            opts = None
            logger.error("Could not identify source from  message:{}\n"
                         .format(str(message)))
        return opts

    def get_name(self):
        """
        Return the name the Lego recognizes from the help function.

        :return: a string
        """
        return '?'

    def get_help(self):
        """
        Return a helpstring for the function.

        :return: a string
        """
        return ''

    def on_failure(self, exception_type, exception_value, traceback):
        ref = ActorRegistry.get_by_urn(self.actor_urn)
        logger.exception('Lego crashed: ' + str(ref))
        logger.exception(exception_type)
        logger.exception(exception_value)



# Legobot
# Copyright (C) 2016 Brenton Briggs, Kevin McCabe, and Drew Bronson

'''
.. module:: Legobot.Connectors.Slack
  :synopsis: A backend for connecting Legobot to Slack

.. moduleauthor:: Bren Briggs

Module lovingly built with inspiration from slackhq's RtmBot
'''

import logging
import re
import sys
import threading
import time

from slackclient import SlackClient

from Legobot.Lego import Lego
from Legobot.Message import Message, Metadata
from Legobot.Utilities import Utilities

logger = logging.getLogger(__name__)

# I forgot why this is here, but it's very necessary.
sys.dont_write_bytecode = True


class RtmBot(threading.Thread, object):
    """
    Creates a Slack bot using the Slackclient RTM API

    Attributes:
        baseplate: The parent pykka actor
        token: The Slack API token
        last_ping: Timestamp of last keepalive ping
        actor_urn: URN of the Pykka actor the RtmBot runs in
        supported_events: dict of Slack RTM API events we care about
        slack_client: initialized RTM client
    """

    def __init__(self, baseplate, token, actor_urn, reconnect=True,
                 *args, **kwargs):
        """Initializes RtmBot

        Args:
            baseplate (Legobot.Lego): The parent Pykka actor.
                Typically passed in fromLegobot.Connectors.Slack.Slack
            token (string): Slack API token
            actor_urn (string): URN of Pykka actor launching RtmBot
            *args: Variable length argument list.
            **kwargs: Arbitrary keyword arguments.

        Returns:
            Returns an instance of RtmBot which connects to Slack.
        """
        self.baseplate = baseplate
        self.token = token
        self.last_ping = 0
        self.actor_urn = actor_urn
        self.reconnect = reconnect
        self.auto_reconnect = True
        # 'event':'method'
        self.supported_events = {'message': self.on_message}
        self.user_map = {}
        self.slack_client = SlackClient(self.token)
        threading.Thread.__init__(self)

    def connect(self):
        self.slack_client.rtm_connect(reconnect=self.reconnect,
                                      auto_reconnect=self.auto_reconnect)

    def on_message(self, event):
        '''Runs when a message event is received

        Args:
            event: RTM API event.

        Returns:
            Legobot.messge
        '''

        metadata = self._parse_metadata(event)
        message = Message(text=metadata['text'],
                          metadata=metadata).__dict__
        if message.get('text'):
            message['text'] = self.find_and_replace_userids(message['text'])
            message['text'] = self.find_and_replace_channel_refs(
                message['text']
            )
        return message

    def run(self):
        '''Extends the run() method of threading.Thread
        '''

        self.connect()
        while True:
            for event in self.slack_client.rtm_read():
                logger.debug(event)
                if 'type' in event and event['type'] in self.supported_events:
                    event_type = event['type']
                    dispatcher = self.supported_events[event_type]
                    message = dispatcher(event)
                    logger.debug(message)
                    self.baseplate.tell(message)
            self.keepalive()
            time.sleep(0.1)
        return

    def find_and_replace_userids(self, text):
        '''Finds occurrences of Slack userids and attempts to replace them with
           display names.

        Args:
            text (string): The message text
        Returns:
            string: The message text with userids replaced.
        '''

        match = True
        pattern = re.compile('<@([A-Z0-9]{9})>')
        while match:
            match = pattern.search(text)
            if match:
                name = self.get_user_display_name(match.group(1))
                text = re.sub(re.compile(match.group(0)), '@' + name, text)

        return text

    def find_and_replace_channel_refs(self, text):
        '''Find occurrences of Slack channel referenfces and attempts to
           replace them with just channel names.

        Args:
            text (string): The message text
        Returns:
            string: The message text with channel references replaced.
        '''

        match = True
        pattern = re.compile('<#([A-Z0-9]{9})\|([A-Za-z0-9-]+)>')
        while match:
            match = pattern.search(text)
            if match:
                text = text.replace(match.group(0), '#' + match.group(2))

        return text

    def get_channel_id_by_name(self, name):
        channels = self.get_channels(condensed=True)
        if not channels:
            return name

        if name.startswith('#'):
            name = name[1:]

        channels_transform = {channel.get('name'): channel.get('id')
                              for channel in channels}
        return channels_transform.get(name)

    def get_user_id_by_name(self, name):
        users = self.get_users(condensed=True)
        if not users:
            return name

        if name.startswith('@'):
            name = name[1:]

        users_transform = {}
        for user in users:
            users_transform[user.get('name')] = user.get('id')
            users_transform[user.get('display_name')] = user.get('id')

        if name in users_transform.keys():
            return users_transform[name]
        else:
            return name

    def get_channels(self, condensed=False):
        '''Grabs all channels in the slack team

        Args:
            condensed (bool): if true triggers list condensing functionality

        Returns:
            dic: Dict of channels in Slack team.
                See also: https://api.slack.com/methods/channels.list
        '''

        channel_list = self.slack_client.api_call('channels.list')
        if not channel_list.get('ok'):
            return None

        if condensed:
            channels = [{'id': item.get('id'), 'name': item.get('name')}
                        for item in channel_list.get('channels')]
            return channels
        else:
            return channel_list

    def get_users(self, condensed=False):
        '''Grabs all users in the slack team

        This should should only be used for getting list of all users. Do not
        use it for searching users. Use get_user_info instead.

        Args:
            condensed (bool): if true triggers list condensing functionality

        Returns:
            dict: Dict of users in Slack team.
                See also: https://api.slack.com/methods/users.list
        '''

        user_list = self.slack_client.api_call('users.list')
        if not user_list.get('ok'):
            return None
        if condensed:
            users = [{'id': item.get('id'), 'name': item.get('name'),
                     'display_name': item.get('profile').get('display_name')}
                     for item in user_list.get('members')]
            return users
        else:
            return user_list

    def get_user_display_name(self, userid):
        '''Given a Slack userid, grabs user display_name from api.

        Args:
            userid (string): the user id of the user being queried
        Returns:
            dict: a dictionary of the api response
        '''

        user_info = self.slack_client.api_call('users.info', user=userid)
        if user_info.get('ok'):
            user = user_info.get('user')
            if user.get('profile'):
                return user.get('profile').get('display_name')
            else:
                return user.get('name')
        else:
            return userid

    def get_dm_channel(self, userid):
        '''Perform a lookup of users to resolve a userid to a DM channel

        Args:
            userid (string): Slack userid to lookup.

        Returns:
            string: DM channel ID of user
        '''

        dm_open = self.slack_client.api_call('im.open', user=userid)
        return dm_open['channel']['id']

    def post_attachment(self, attachment):
        self.slack_client.api_call('chat.postMessage', **attachment)

    def get_username(self, userid):
        '''Perform a lookup of users to resolve a userid to a username

        Args:
            userid (string): Slack userid to lookup.

        Returns:
            string: Human-friendly name of the user
        '''

        username = self.user_map.get(userid)
        if not username:
            users = self.get_users()
            if users:
                members = {
                    m['id']: m['name']
                    for m in users.get('members', [{}])
                    if m.get('id')
                    and m.get('name')
                }
                if members:
                    self.user_map.update(members)

                username = self.user_map.get(userid, userid)

        return username

    def get_userid_from_botid(self, botid):
        '''Perform a lookup of bots.info to resolve a botid to a userid

        Args:
            botid (string): Slack botid to lookup.
        Returns:
            string: userid value
        '''
        botinfo = self.slack_client.api_call('bots.info', bot=botid)
        if botinfo['ok'] is True:
            return botinfo['bot'].get('user_id')
        else:
            return botid

    def _parse_metadata(self, message):
        '''Parse incoming messages to build metadata dict
        Lots of 'if' statements. It sucks, I know.

        Args:
            message (dict): JSON dump of message sent from Slack

        Returns:
            Legobot.Metadata
        '''

        # Try to handle all the fields of events we care about.
        metadata = Metadata(source=self.actor_urn).__dict__
        metadata['thread_ts'] = message.get('thread_ts')
        if 'presence' in message:
            metadata['presence'] = message['presence']

        if 'text' in message:
            metadata['text'] = message['text']
        elif 'previous_message' in message:
            # Try to handle slack links
            if 'text' in message['previous_message']:
                metadata['text'] = message['previous_message']['text']
            else:
                metadata['text'] = None
        else:
            metadata['text'] = None

        if 'user' in message:
            metadata['source_user'] = message['user']
        elif 'bot_id' in message:
            metadata['source_user'] = self.get_userid_from_botid(
                                      message['bot_id'])
        elif 'message' in message and 'user' in message['message']:
            metadata['source_user'] = message['message']['user']
        else:
            metadata['source_user'] = None

        metadata['user_id'] = metadata['source_user']
        metadata['display_name'] = self.get_username(metadata['source_user'])

        if 'channel' in message:
            metadata['source_channel'] = message['channel']
            # Slack starts DM channel IDs with "D"
            if message['channel'].startswith('D'):
                metadata['is_private_message'] = True
            else:
                metadata['is_private_message'] = False

        metadata['source_connector'] = 'slack'

        return metadata

    def keepalive(self):
        '''Sends a keepalive to Slack
        '''
        # hardcode the interval to 3 seconds
        now = int(time.time())
        if now > self.last_ping + 3:
            self.slack_client.server.ping()
            self.last_ping = now


class Slack(Lego):
    '''Lego that builds and connects Legobot.Connectors.Slack.RtmBot

    Args:
        baseplate (Legobot.Lego): baseplate/parent lego (Pykka Actor)
        lock (threading.Lock: thread lock created in your bot script.
            All legos should share the same lock.
        *args: Variable length argument list.
        **kwargs: Arbitrary keyword arguments.
    '''

    def __init__(self, baseplate, lock, *args, **kwargs):
        super().__init__(baseplate, lock)
        self.botThread = RtmBot(baseplate, actor_urn=self.actor_urn,
                                *args, **kwargs)

    def on_start(self):
        '''Extends pykka's on_start method to launch this as an actor
        '''

        self.botThread.start()

    def listening_for(self, message):
        '''Describe what this should listen for (hint: everything)
        Extends Legobot.Lego.listening_for()

        Args:
            message (Legobot.Message): Message to handle

        Returns:
            bool: True if lego is interested in the message.
        '''

        return str(self.botThread) != str(message['metadata']['source'])

    def build_attachment(self, text, target, attachment, thread):
        '''Builds a slack attachment.

        Args:
            message (Legobot.Message): message w/ metadata to send.

        Returns:
            attachment (dict): attachment data.
        '''

        attachment = {
            'as_user': True,
            'text': text,
            'channel': target,
            'attachments': [
                {
                    'fallback': text,
                    'image_url': attachment
                }
            ]
        }
        if thread:
            attachment['thread_ts'] = thread
        return attachment

    def handle(self, message):
        '''Attempts to send a message to the specified destination in Slack.
        Extends Legobot.Lego.handle()

        Args:
            message (Legobot.Message): message w/ metadata to send.
        '''

        logger.debug(message)
        if Utilities.isNotEmpty(message['metadata']['opts']):
            target = message['metadata']['opts']['target']
            thread = message['metadata']['opts'].get('thread')
            # pattern = re.compile('@([a-zA-Z0-9._-]+)')
            pattern = re.compile('^@([a-zA-Z0-9._-]+)|\s@([a-zA-Z0-9._-]+)')
            matches = re.findall(pattern, message['text'])
            matches = set(matches)
            logger.debug('MATCHES!!!!   {}'.format(matches))
            for match in matches:
                if isinstance(match, tuple):
                    if match[0] != '':
                        match = match[0]
                    else:
                        match = match[1]
                if not match.startswith('@'):
                    match = '@' + match
                message['text'] = message['text'].replace(
                    match,
                    '<{}>'.format(match)
                )

            pattern = re.compile('#([A-Za-z0-9-]+)')
            matches = re.findall(pattern, message['text'])
            matches = set(matches)
            for match in matches:
                channel_id = self.botThread.get_channel_id_by_name(match)
                if channel_id:
                    message['text'] = message['text'].replace(
                        '#' + match,
                        '<#{}|{}>'.format(
                            channel_id,
                            match
                        )
                    )

            if (message['text'].find('<<@') != -1
                    or message['text'].find('<<#') != -1):
                message['text'] = message['text'].replace('<<', '<')
                message['text'] = message['text'].replace('>>', '>')

            if target.startswith('U'):
                target = self.botThread.get_dm_channel(target)
            attachment = message['metadata']['opts'].get('attachment')
            if attachment:
                text = message['metadata']['opts'].get('fallback')
                attachment = self.build_attachment(
                    text, target, attachment, thread)
                self.botThread.post_attachment(attachment)
            else:
                self.botThread.slack_client.rtm_send_message(
                    target, message['text'], thread=thread)

    @staticmethod
    def get_name():
        '''Called by built-in !help lego
        Returns name of Lego. Returns none because this is
        a non-interactive Lego
        '''

        return None



# Legobot
# Copyright (C) 2016 Brenton Briggs, Kevin McCabe, and Drew Bronson

import threading
import unittest

from Legobot.Lego import Lego
from Legobot.Legos.TestingConnector import TestingConnector
from Legobot.Message import Message, Metadata


class TestTestingConnector(unittest.TestCase):
    def test_init(self):
        testing_connector = self._make_testing_connector('blah.tmp')
        assert testing_connector.temp_file == 'blah.tmp'  # nosec

    def test_default_parameters(self):
        testing_connector = self._make_testing_connector()
        assert testing_connector.temp_file == 'testing_file.tmp'  # nosec

    def test_listening_for(self):
        testing_connector = self._make_testing_connector()
        assert testing_connector.listening_for(self._make_message())  # nosec

    def test_handle(self):
        testing_connector = self._make_testing_connector()
        message = self._make_message()
        testing_connector.handle(message)
        with open(testing_connector.temp_file, mode='r') as f:
            text = f.read()
        assert text == str(message)  # nosec

    @staticmethod
    def _make_testing_connector(temp_file=None):
        lock = threading.Lock()
        baseplate = Lego(None, lock)
        if temp_file is None:
            testing_connector = TestingConnector(baseplate, lock)
        else:
            testing_connector = TestingConnector(baseplate, lock, temp_file)
        return testing_connector  # nosec

    @staticmethod
    def _make_message():
        source = Lego(None, threading.Lock())
        metadata = Metadata(source)
        message = Message('blah', metadata)
        return message



# Legobot
# Copyright (C) 2016 Brenton Briggs, Kevin McCabe, and Drew Bronson

# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.

# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.



from websocket import create_connection
import json
import logging
import requests
import sys
import threading
import time

from Legobot.Lego import Lego
from Legobot.Message import Message, Metadata
from Legobot.Utilities import Utilities

__author__ = "Bren Briggs (fraq)"
__copyright__ = "Copyright 2017, Legobot"

logger = logging.getLogger(__name__)


class Heartbeat(threading.Thread, object):
    """
    Manage heartbeat connection to Discord

    Attributes:
        ws: Websocket connection to Discord
        interval (float): time between heartbeat pings to the server

    """

    def __init__(self, ws, interval):
        self.ws = ws
        self.interval = float(interval) / 1000.0
        self.sequence = 0
        logger.debug("Heartbeat handler initialized \
                     with interval of {}".format(self.interval))
        threading.Thread.__init__(self)

    def send(self, ws, seq):
        """
        Sends heartbeat message to Discord

        Attributes:
            ws: Websocket connection to discord
            seq: Sequence number of heartbeat
        """

        payload = {'op': 1, 'd': seq}
        payload = json.dumps(payload)
        logger.debug("Sending heartbeat with payload {}".format(payload))
        ws.send(payload)
        return

    def update_sequence(self, seq):
        """

        Attributes:
            seq (int): Sequence number of heartbeat
        """

        self.sequence = seq
        logger.debug("Heartbeat sequence updated to {}".format(self.sequence))
        return

    def run(self):
        while True:
            self.send(self.ws, self.sequence)
            time.sleep(self.interval)


class DiscoBot(threading.Thread, object):
    """
    Relays messages between Discord and the Legobot baseplate.
    Uses the REST API and Websockets. No plans to use async.

    Attibutes:
        actor_urn: URN (unique ID) of actor that initialized the bot
        baseplate: Parent lego/actor
        headers: HTTP headers use for calls to Discord REST endpoints
        rest_baseurl: Base URL to the REST API
        token: Discord bot token
        ws: Websocket connected to Discord
    """

    def __init__(self, baseplate, token, actor_urn, *args, **kwargs):
        """
        Initialize DiscoBot

        Args:
            baseplate (Legobot.Lego): The parent Pykka actor.
                Typically passed in from Legobot.Connectors.Discord.Discord
            token (string): Discord bot token
            actor_urn (string): URN of Pykka actor launching DiscoBot
            *args: Variable length argument list.
            **kwargs: Arbitrary keyword arguments.
        """

        self.baseplate = baseplate
        self.rest_baseurl = 'https://discordapp.com/api/v6'
        self.token = token
        self.headers = {"Authorization": "Bot {}".format(token),
                        "User-Agent": "Legobot",
                        "Content-Type": "application/json"}
        self.actor_urn = actor_urn
        self.ws = None
        threading.Thread.__init__(self)

    def connect(self):
        return create_connection(self.get_ws_url())

    def create_message(self, channel_id, text):
        """
        Sends a message to a Discord channel or user via REST API

        Args:
            channel_id (string): ID of destingation Discord channel
            text (string): Content of message
        """

        baseurl = self.rest_baseurl + \
            '/channels/{}/messages'.format(channel_id)
        requests.post(baseurl,
                      headers=self.headers,
                      data=json.dumps({'content': text}))

    def identify(self, token):
        """
        Identifies to the websocket endpoint

        Args:
            token (string): Discord bot token
        """

        payload = {
            'op': 2,
            'd': {
                'token': self.token,
                'properties': {
                    '$os': sys.platform,
                    '$browser': 'legobot',
                    '$device': 'legobot'
                },
                'compress': False,
                'large_threshold': 250
            }
        }
        payload['d']['synced_guilds'] = []
        logger.info("Identifying with the following message: \
                    {}".format(payload))
        self.ws.send(json.dumps(payload))
        return

    def on_hello(self, message):
        """
        Runs on a hello event from websocket connection

        Args:
            message (dict): Full message from Discord websocket connection"
        """

        logger.info("Got a hello")
        self.identify(self.token)
        self.heartbeat_thread = Heartbeat(self.ws,
                                          message['d']['heartbeat_interval'])
        self.heartbeat_thread.start()
        return

    def on_heartbeat(self, message):
        """
        Runs on a heartbeat event from websocket connection

        Args:
            message (dict): Full message from Discord websocket connection"
        """

        logger.info("Got a heartbeat")
        logger.info("Heartbeat message: {}".format(message))
        self.heartbeat_thread.update_sequence(message['d'])
        return

    def on_message(self, message):
        """
        Runs on a create_message event from websocket connection

        Args:
            message (dict): Full message from Discord websocket connection"
        """

        if 'content' in message['d']:
            metadata = self._parse_metadata(message)
            message = Message(text=message['d']['content'],
                              metadata=metadata).__dict__
            logger.debug(message)
            self.baseplate.tell(message)

    def _parse_metadata(self, message):
        """
        Sets metadata in Legobot message

        Args:
            message (dict): Full message from Discord websocket connection"

        Returns:
            Legobot.Metadata
        """

        metadata = Metadata(source=self.actor_urn).__dict__
        if 'author' in message['d']:
            metadata['source_user'] = message['d']['author']['username']
        else:
            metadata['source_user'] = None
        if 'channel_id' in message['d']:
            metadata['source_channel'] = message['d']['channel_id']
        else:
            metadata['source_channel'] = None
        metadata['user_id'] = metadata['source_user']
        metadata['display_name'] = metadata['source_user']

        metadata['source_connector'] = 'discord'

        return metadata

    @staticmethod
    def get_ws_url():
        url = requests.get(
            'https://discordapp.com/api/v6/gateway').json()['url']
        return url + '?v=6&encoding=json'

    def handle(self, message):
        """
        Dispatches messages to appropriate handler based on opcode

        Args:
            message (dict): Full message from Discord websocket connection
        """

        opcode = message['op']
        if opcode == 10:
            self.on_hello(message)
        elif opcode == 11:
            self.on_heartbeat(message)
        elif opcode == 0:
            self.on_message(message)
        else:
            logger.debug("Not a message we handle: OPCODE {}".format(opcode))
        return

    def run(self):
        """
        Overrides run method of threading.Thread.
        Called by DiscoBot.start(), inherited from threading.Thread
        """

        self.ws = self.connect()
        while True:
            try:
                data = json.loads(self.ws.recv())
                self.handle(data)
            except json.decoder.JSONDecodeError as e:
                logger.fatal("No data on socket...")


class Discord(Lego):
    '''
    Lego that builds and connects Legobot.Connectors.Discord.DiscoBot

    Args:
        baseplate (Legobot.Lego): baseplate/parent lego (Pykka Actor)
        lock (threading.Lock: thread lock created in your bot script.
            All legos should share the same lock.
        *args: Variable length argument list.
        **kwargs: Arbitrary keyword arguments.
    '''

    def __init__(self, baseplate, lock, *args, **kwargs):
        super().__init__(baseplate, lock)
        self.botThread = DiscoBot(baseplate, actor_urn=self.actor_urn,
                                  *args, **kwargs)

    def on_start(self):
        '''
        Extends pykka's on_start method to launch this as an actor
        '''

        self.botThread.start()

    def listening_for(self, message):
        '''
        Describe what this should listen for
        (hint: everything not from ourself)
        Extends Legobot.Lego.listening_for()

        Args:
            message (Legobot.Message): Message to handle

        Returns:
            bool: True if lego is interested in the message.
        '''

        return str(self.actor_urn) != str(message['metadata']['source'])

    def handle(self, message):
        '''
        Attempts to send a message to the specified destination in Discord.
        Extends Legobot.Lego.handle()

        Args:
            message (Legobot.Message): message w/ metadata to send.
        '''

        logger.debug(message)
        if Utilities.isNotEmpty(message['metadata']['opts']):
            target = message['metadata']['opts']['target']
            self.botThread.create_message(target, message['text'])

    @staticmethod
    def get_name():
        '''
        Called by built-in !help lego

        Returns name of Lego. Returns None because this is
        a non-interactive Lego
        '''

        return None
